# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
import subprocess
import time
import re

BUILD_METADATA_FILE = "rust-update-set.build.metadata"
DIVIDING_LINE = "="*150

class Builder():
    def __init__(self, checkout_metadata, build_metadata, project):
        self.dependencies_data = checkout_metadata["dependencies"]
        self.pkgs_data = checkout_metadata["checkout"]
        self.build_metadata = build_metadata
        self.project = project

    def analyse_build_order(self):
        """
        Analyze dependecies and specify an order to build packages in parallel
        DAG representation: dependent -> dependency
        Start with the node with no outgoing edges (without dependencies)
        """
        
        incoming_edges = {}
        outgoing_edges = {}
        crates = set()
        for dependent in self.dependencies_data:
            crates.add(dependent)
            dependencies = self.dependencies_data[dependent]
            for dependency in dependencies:
                incoming_edges[dependency] = incoming_edges.get(dependency, set())
                incoming_edges[dependency].add(dependent)
                outgoing_edges[dependent] = outgoing_edges.get(dependent, set())
                outgoing_edges[dependent].add(dependency)
        orders = []
        cur_order = []
        for crate in self.dependencies_data:
            if crate not in outgoing_edges:
                cur_order.append(crate)
                crates.remove(crate)
        while len(crates) > 0:
            orders.append(cur_order)
            next_order = []
            for crate in cur_order:
                for dependent in incoming_edges.get(crate, set()):
                    outgoing_edges[dependent].remove(crate)
                    if len(outgoing_edges[dependent]) == 0:
                        next_order.append(dependent)
                        crates.remove(dependent)
            cur_order = next_order
        orders.append(cur_order)
        return orders

    def project_exists(self):
        """ Check if the copr project already exists """
        output = subprocess.run(["copr-cli", "list"], capture_output=True)
        return f"Name: {self.project}" in output.stdout.decode()

    def create_copr_project(self):
        """
        Create a copr project
        FIX: probably need to make chroot as passed in arguments, also enabling fedora-review
        """
        output = subprocess.run(["copr-cli", "create", self.project, "--chroot", "fedora-rawhide-x86_64", "--fedora-review"], capture_output=True)
        if output.returncode == 0:
            print(f"New Copr project {self.project} was successfully created")
        else:
            self.print_cmd_output(output)

    def read_fas_user(self):
        """
        Try to read FAS username
        FIX: would it be safe enough to use fas username to register src.fedoraproject.org repository?
        """
        if os.path.exists(os.path.expanduser("~/.fedora.upn")):
            with open(os.path.expanduser("~/.fedora.upn"), "r") as f:
                return f.read().replace("\n", "")

        klist_line = "Default principal: "
        koutput = subprocess.check_output("klist", universal_newlines=True).split("\n")
        line = [l for l in koutput if l.startswith(klist_line)][0]
        line = line[len(klist_line) :]
        if line.endswith("@FEDORAPROJECT.ORG"):
            print(f"Fetch FAS username: {line[: line.index('@')]}")
            return line[: line.index("@")]

        raise Exception("Could not determine FAS username")

    def initialize_build_metadata(self):
        """ Initialize build metadata after figuring out build order"""
        print(DIVIDING_LINE)
        print("Dependency Order Analyses")
        print(DIVIDING_LINE)
        orders = self.analyse_build_order()
        print("Packages will be built in the following order:")
        for i, order in enumerate(orders):
            print(f"{i+1}. {order}")
        for i, order in enumerate(orders):
            self.build_metadata[str(i)] = {}
            for crate in order:
                self.build_metadata[str(i)][f"rust-{crate}"] = {
                    "type": self.pkgs_data[f"rust-{crate}"]["type"],
                    "status": "new" # new -> registered -> success
                }
        self.dump_metadata()

    def register_existing_pkgs(self):
        """ Register all existing package in copr project with git url """
        print(DIVIDING_LINE)
        print("Copr Packages Registration")
        print(DIVIDING_LINE)
        for i in range(len(self.build_metadata)):
            i = str(i)
            for pkg in self.build_metadata[i]:
                type_, status = self.build_metadata[i][pkg]["type"], self.build_metadata[i][pkg]["status"]
                if type_ == "existing" and status == "new":
                    # if pushing to actual repo, register with actual repo git
                    # if pushing to forked repo, register with forked repo git
                    # by default, register a forked repo with the name from FAS Account
                    if self.pkgs_data[pkg]["push"] == "origin":
                        clone_url = f"https://src.fedoraproject.org/rpms/{pkg}.git"
                    elif self.pkgs_data[pkg]["push"] == "":
                        clone_url = f"https://src.fedoraproject.org/forks/{self.username}/rpms/{pkg}.git"
                    else:
                        clone_url = f"https://src.fedoraproject.org/forks/{self.pkgs_data[pkg]['push']}/rpms/{pkg}.git"
                    output = subprocess.run(["copr-cli", "add-package-scm", "--name", pkg,
                        "--clone-url", clone_url, self.project], capture_output=True)
                    if output.returncode == 0:
                        print(f"Successfully register package {pkg} with {clone_url}")
                        self.build_metadata[i][pkg]["status"] = "registered"
                        self.dump_metadata()
                    else:
                        self.print_cmd_output(output)
                elif status == "registered":
                    print(f"Package {pkg} is already registered")

    def dump_metadata(self):
        """ Dump build related metadata into json file """
        with open(BUILD_METADATA_FILE, "w") as f:
            json.dump(self.build_metadata, f, indent=4)

    def print_cmd_output(self, output):
        """ Print cmd output with warning color """
        out = output.stdout.decode().split("\n")[:-1]
        err = output.stderr.decode().split("\n")[:-1]
        for line in out:
            print('\033[93m'+line+'\033[0m')
        for line in err:
            print('\033[93m'+line+'\033[0m')
        return len(out)+len(err)

    def copr_build(self):
        """
        Start building packages in copr
        1. create build: build package from git url for existing packages and from local srpm for new packages
        2. checking all parallel builds: all success then move to next order, any fails will return
        """
        print(DIVIDING_LINE)
        print("Copr Build")
        print(DIVIDING_LINE)
        for i in range(len(self.build_metadata)):
            # build all pkgs
            i = str(i)
            pkgs_to_build = self.build_metadata[i]
            builds = set()
            for pkg in pkgs_to_build:
                if self.build_metadata[i][pkg]["status"] == "success": continue
                type_ = pkgs_to_build[pkg]["type"]
                if type_ == "existing":
                    output = subprocess.run(["copr-cli", "build-package", self.project, "--name", f"{pkg}", "--nowait"], capture_output=True)
                else:
                    output = subprocess.run(["copr-cli", "build", self.project, f"{pkg}/{pkg}.spec", "--nowait"], capture_output=True)
                if output.returncode != 0:
                    print('\033[93m'+f"Fail to start building {pkg}"+'\033[0m')
                    self.print_cmd_output(output)
                    continue
                build_id = 0
                for line in output.stdout.decode().split("\n"):
                    if line.startswith("Created builds"):
                        build_id = int(line.split(":")[1])
                builds.add(build_id)
                print(f"Start building {pkg} https://copr.fedorainfracloud.org/coprs/{self.copr_username}/{self.project}/build/{build_id}")

            # monitor all builds, if fails return, else move to next set of pkgs to build
            if len(builds) == 0: continue
            success = True
            build_failes = []
            while len(builds) != 0:
                output = subprocess.run(["copr-cli", "monitor", self.project], capture_output=True)
                if len(output.stdout.decode()) == 0: continue
                all_builds = json.loads(output.stdout.decode())
                for build_data in all_builds:
                    name, build_id, status = build_data["name"], build_data["build_id"], build_data["state"]
                    if build_id in builds:
                        if status == "succeeded":
                            builds.remove(build_id)
                            self.build_metadata[i][name]["status"] = "success"
                            self.dump_metadata()
                            print("\033[32m"+f"Build for {name} succeed: https://copr.fedorainfracloud.org/coprs/{self.copr_username}/{self.project}/build/{build_id}"+"\033[0m")
                        if status == "failed":
                            builds.remove(build_id)
                            success = False
                            build_failes.append((name, build_id))
                time.sleep(10)
            print("-"*150)
            if not success:
                print(f"The following build failes, check on https://copr.fedorainfracloud.org/coprs/{self.copr_username}/{self.project}/builds")
                for build in build_failes:
                    print("\033[93m"+f"      Name: {build[0]}  Build ID: {build[1]}"+"\033[0m")
                return False

        return True

    def build(self):
        """ Entry for building all packages in Copr """
        # check copr config
        output = subprocess.run(["copr-cli", "whoami"], capture_output=True)
        whoami = output.stdout.decode()
        if "None" in whoami:
            print("Cannot read Copr config, please add your API Key in ~/.config/copr.")
            print("Log into the Copr web interface and get your API token at the https://copr.fedorainfracloud.org/api")
            return
        self.copr_username = whoami.split("\n")[0]

        # check if project exists
        exist = self.project_exists()
        if exist:
            print(f"Copr project {self.project} already exists")
        else:
            self.create_copr_project()

        # get FAS username for registering existing project
        try:
            self.username = self.read_fas_user()
        except Exception:
            self.username = input("Please input your FAS username: ")

        # register packages
        if self.build_metadata == {}:
            self.initialize_build_metadata()
        self.register_existing_pkgs()

        # build
        if self.copr_build():
            print(f"All packages are built successfully, check the result at https://copr.fedorainfracloud.org/coprs/{self.copr_username}/{self.project}/packages")

    def get_chain_build_order(self):
        """ Get checkout directory and build order e.g. 'A B : C : D : '"""
        orders = self.analyse_build_order()
        cmd = []
        directory = sorted(orders[-1])[0]
        for i in range(len(orders)):
            for crate in orders[i]:
                if crate != directory:
                    cmd.append(f"rpms/rust-{crate}")
            if i!= len(orders)-1:
                cmd.append(":")
        return f"rust-{directory}", cmd

    def fetch_side_tag(self, output):
        """ Parse side tag from fedpkg request-side-tag output """
        pattern = r"Side tag '[\w-]+'"
        match = re.search(pattern, output)
        if match is None:
            return ""
        return match.group(0)[10:-1]
    
    def chain_build(self, dry_run=False, branch=None):
        """ 
        Chain-build all packages using fedpkg side-tag and chain-build 
        If no branch is specified, will use the current branch
        """
        # checkout last directory, request side-tag and start chain-build
        directory, cmd = self.get_chain_build_order()
        print(f"Check out directory {directory}")
        os.chdir(f"{os.getcwd()}/{directory}")
        if branch is not None:
            output = subprocess.run(["git", "switch", branch], capture_output=True)
            if output.returncode == 0:
                print(f"Successfully switch to branch {branch}")
            else:
                print(f"Failed to switch to branch {branch}")
                self.print_cmd_output(output)
        output = subprocess.run(["fedpkg", "request-side-tag"], capture_output=True)
        if output.returncode != 0:
            print("Failed to run 'fedpkg request-side-tag")
            self.print_cmd_output(output)
            return
        side_tag = self.fetch_side_tag(output.stdout.decode())
        if side_tag == "":
            print("Failed to fetch side-tag from 'fedpkg request-side-tag' output:")
            self.print_cmd_output(output)
            return
        cmd = ["fedpkg", "chain-build", f"--target={side_tag}", *cmd]
        if dry_run:
            print("Command to be executed: ")
            print(" ".join(cmd))
        else:
            subprocess.run(cmd)