# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import requests
import configparser
import tempfile
import tarfile
import os
import subprocess
import re
from typing import List
from bodhi.client.bindings import BodhiClient
from rust2rpm.core.metadata import CargoSemVer, Metadata

CRATE_API_URL = "https://crates.io/api/v1/"


class DChecker:
    def __init__(self, bodhi_client=None):
        self.package_version_cache = {}
        self.bodhi_client = bodhi_client or BodhiClient()
        # (crate, version) -> {dep.name: (dep.req, dep.nor_vers, dep.features)}
        self.dependecies_from_crateio_cache = {}

    def parse_compose_req(self, req):
        """Parse compose req like '>=0.1.0, <0.2' to normalized reqs [(kind, version), (kind, version)]"""
        reqs = req.split(",")
        parsed_reqs = [
            CargoSemVer.parse(reqs[0].strip()),
            CargoSemVer.parse(reqs[1].strip()),
        ]
        return [
            (parsed_reqs[0].kind, CargoSemVer.coerce(parsed_reqs[0].version)),
            (parsed_reqs[1].kind, CargoSemVer.coerce(parsed_reqs[1].version)),
        ]

    def get_versions(self, crate):
        """Get all existing version for a crate from crate.io"""
        url = requests.compat.urljoin(CRATE_API_URL, f"crates/{crate}/versions")
        req = requests.get(url, headers={"User-Agent": "rust-update-set"})
        req.raise_for_status()
        versions = req.json()["versions"]
        return versions

    def get_latest_crate_version(self, crate):
        """Get the latest version of crate from crate.io"""
        versions = self.get_versions(crate)
        version = next(version["num"] for version in versions if not version["yanked"])
        return version

    def get_cloest_crate_version(self, crate, version):
        """
        Get the cloest version of a specific version
        Example: return 6.0.2 for a dependency with ^6.0 if 6.0.0 and 6.0.1 doesn't exist
        """
        versions = self.get_versions(crate)
        latest_version = None
        compare_version = CargoSemVer.parse_version(version)
        for md in versions:
            cur_version = CargoSemVer.parse_version(md["num"])
            # ignore prerelease version like cursive 0.17.0-alpha.0
            if cur_version.pre_release:
                continue
            # TODO: simplify once https://pagure.io/fedora-rust/rust2rpm/pull-request/229
            # is merged and released
            if not md["yanked"] and (
                CargoSemVer.eval_(cur_version, CargoSemVer.KIND_GT, compare_version)
                or CargoSemVer.eval_(
                    cur_version, CargoSemVer.KIND_SHORTEQ, compare_version
                )
            ):
                latest_version = cur_version
            if not md["yanked"] and CargoSemVer.eval_(
                cur_version, "<", compare_version
            ):
                break
        return CargoSemVer.unparse_version(latest_version)

    def get_unwanted_features(self) -> List[str]:
        """
        Extract unwanted features from rust2rpm.conf
        FIX: now only DEFAULT setting will be considered, probably add others like [fedora]
        """
        conf = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation()
        )
        confs = conf.read([".rust2rpm.conf", "_rust2rpm.conf", "rust2rpm.conf"])
        unwanted_features = conf["DEFAULT"].get("unwanted-features", None)
        if unwanted_features is not None:
            return [
                feature for feature in unwanted_features.split("\n") if feature != ""
            ]
        return None

    def get_dependencies_from_local(self, crate, version):
        """Get all dependencies from local .crate source file"""
        unwanted_features = self.get_unwanted_features()
        with self.files_from_crate(f"{crate}-{version}.crate", crate, version) as toml:
            metadata = Metadata.from_file(toml)[0]
        # ignore unwanted features
        dependencies = set(
            [
                (dependency.name, dependency.req)
                for dependency in metadata.dependencies[None][1]
            ]
        )
        for feature in metadata.dependencies:
            if feature is None or (
                unwanted_features is not None and feature in unwanted_features
            ):
                continue
            for dependency in metadata.dependencies[feature][1]:
                dependencies.add((dependency.name, dependency.req))
        for dependency in metadata.dev_dependencies:
            dependencies.add((dependency.name, dependency.req))
        return list(dependencies)

    def reporequery_package_version(self, crate):
        """Get the lateset packaged version of a crate"""
        cmd = f"sudo dnf repoquery --disablerepo=\* --enablerepo=rawhide,rawhide-source rust-{crate} --nvr"
        run = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        output, err = run.communicate()
        if run.returncode != 0:
            print(f"Fail to execute command {cmd}")
            print(err.decode())
            return -1
        for nvr in output.decode().split("\n"):
            if nvr.startswith(f"rust-{crate}"):
                name, version = self.fetch_name_version_from_nvr(nvr)
                return version
        return -1

    def check_exist(self, crate):
        """Check if a crate is already packaged, if so, return the latest version"""
        if crate in self.package_version_cache:
            latest_version = self.package_version_cache[crate]
            return latest_version != -1, latest_version
        version = self.reporequery_package_version(crate)
        if version == -1:
            return False, -1  # should not happen
        self.package_version_cache[crate] = version
        return True, version

    def eval(self, v1, op, v2):
        """Rewrite CargoSemVer.eval_() function to compare two versions"""
        v1 = (int(v1.major), int(v1.minor), int(v1.patch))
        v2 = (int(v2.major), int(v2.minor), int(v2.patch))
        if op == "<=":
            return v1 <= v2
        elif op == ">=":
            return v1 >= v2
        elif op == "=":
            return v1 == v2
        elif op == ">":
            return v1 > v2
        elif op == "<":
            return v1 < v2
        else:
            raise ValueError(f"Cannot evaluate operator: {op}")

    def get_missing_old_dependencies(self, crate, version):
        """Filter all dependencies and return missing dependencies and old dependencies with old version"""
        all_dependencies = self.get_dependencies_from_local(crate, version)
        dependencies = []
        for dependency in all_dependencies:
            name, req = dependency[0], dependency[1]
            if "," in req:
                normalized_vers = self.parse_compose_req(req)
            else:
                normalized_vers = CargoSemVer.normalize(CargoSemVer.parse(req))
            req_version = normalized_vers[0][1]  # lower bound of required versions
            exist, latest_version = self.check_exist(name)
            if exist:
                latest_version_no_tilde = latest_version.replace("~", "-")
                latest_version_obj = CargoSemVer.parse_version(latest_version_no_tilde)
                if not self.eval(latest_version_obj, ">=", req_version):
                    updated_version = self.get_cloest_crate_version(
                        name,
                        str(req_version.major)
                        + "."
                        + str(req_version.minor)
                        + "."
                        + str(req_version.patch),
                    )
                    dependencies.append(
                        (name, latest_version, updated_version, "existing")
                    )
            if not exist:
                updated_version = self.get_cloest_crate_version(
                    name,
                    str(req_version.major)
                    + "."
                    + str(req_version.minor)
                    + "."
                    + str(req_version.patch),
                )
                dependencies.append((name, -1, updated_version, "new"))
        return dependencies

    @contextlib.contextmanager
    def files_from_crate(self, cratef, crate, version):
        """Extract Cargo.toml from source .crate, if a fix-metadata.diff is found, patch to Cargo.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            target_dir = f"{tmpdir}/"
            with tarfile.open(cratef, "r") as archive:
                for n in archive.getnames():
                    if not os.path.abspath(os.path.join(target_dir, n)).startswith(
                        target_dir
                    ):
                        raise Exception("Unsafe filenames!")
                archive.extractall(target_dir)
            toml_relpath = f"{crate}-{version}/Cargo.toml"
            toml = f"{tmpdir}/{toml_relpath}"
            # rust2rpm v22 will generate auto diff to strip all foreign dependencies
            if os.path.exists(f"{crate}-fix-metadata-auto.diff"):
                with open(f"{crate}-fix-metadata-auto.diff") as diff:
                    output = subprocess.run(
                        ["patch", toml], stdin=diff, capture_output=True
                    )
            if os.path.exists(f"{crate}-fix-metadata.diff"):
                with open(f"{crate}-fix-metadata.diff") as diff:
                    output = subprocess.run(
                        ["patch", toml], stdin=diff, capture_output=True
                    )
            yield toml

    def is_compatible(self, normalized_vers, updated_version):
        """Check if the updated version is compatible with required versions"""
        for op, version in normalized_vers:
            if not self.eval(updated_version, op, version):
                return False
        return True

    def check_compatibility(self, pkgs):
        """
        Dependency check for each package, check all dependents and see if the dependency version is compatibile
        Return compat packges found during repoquery and packages which breaks others
        """
        compat_pkgs_found = set()
        pkgs_incompatible = []
        for pkg in pkgs:
            if pkgs[pkg]["type"] != "existing":
                continue
            crate, version = pkg[5:], pkgs[pkg]["version"]
            print(f"Checking compatibility for updating rust-{crate} to {version}")
            dependents = self.reporequery_whatrequrires(crate)
            all_compat = True
            # find all compat versions
            compat_versions = set()
            for name, ver in dependents:
                if self.is_compat_pkg(name):
                    name_fixed = self.fix_compat_pkg_name(name, ver)
                    # if the dependent is a compat version of this crate, ignore it
                    if name_fixed == crate:
                        compat_pkgs_found.add((f"rust-{name}", ver))
                        compat_versions.add(ver)
            for name, ver in dependents:
                # if the dependent is already in the update set, ignore it
                if name == crate or f"rust-{name}" in pkgs:
                    continue
                dependencies = self.get_dependencies_from_repoquery(name)
                if crate not in dependencies:
                    # should not happen
                    continue
                parsed_version = CargoSemVer.coerce(CargoSemVer.parse_version(version))
                normalized_reqs = dependencies[crate][1]
                is_compat = self.is_compatible(normalized_reqs, parsed_version)
                if not is_compat:
                    for ver in compat_versions:
                        if self.is_compatible(
                            normalized_reqs,
                            CargoSemVer.coerce(CargoSemVer.parse_version(ver)),
                        ):
                            is_compat = True
                            break
                if not is_compat:
                    all_compat = False
                    print(
                        "\033[93m"
                        + f"      rust-{name} {ver} will be broken since it requires dependency {dependencies[crate][0]}"
                        + "\033[0m"
                    )
            if all_compat:
                print("\033[92m" + f"      Safe to update." + "\033[0m")
            else:
                pkgs_incompatible.append(pkg)
        return compat_pkgs_found, pkgs_incompatible

    def fetch_new_pkg_from_bug(self, title: str):
        """Parse crate name from title of Bugzilla Review Request Bug"""
        pattern = r"Review Request: rust-[\w_-]+"
        match = re.search(pattern, title)
        if match:
            return match.group(0)[21:]
        return ""

    def fetch_name_version_from_nvr(self, nvr: str):
        """Parse name, version from a nvr format"""
        items = nvr.split("-")
        pattern = r"\d+.\d+.\d+"
        for index, item in enumerate(items):
            if re.search(pattern, item):
                return "-".join(items[:index]), item

    def is_compat_pkg(self, name: str) -> bool:
        """Check the name of a package to see if it is a compat crate or real crate"""
        try:
            url = requests.compat.urljoin(CRATE_API_URL, f"crates/{name}")
            req = requests.get(url, headers={"User-Agent": "rust-update-set"})
            req.raise_for_status()
            return False
        except:
            return True

    def reporequery_whatrequrires(self, crate):
        cmd = f"sudo dnf repoquery --disablerepo=\* --enablerepo=rawhide,rawhide-source --whatrequires 'crate({crate})' --whatrequires 'crate({crate}/*)' --nvr"
        run = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        output, err = run.communicate()
        if run.returncode != 0:
            print(f"Fail to execute command {cmd}")
            print(err.decode())
            return set()
        res = set()
        for nvr in output.decode().split("\n"):
            if nvr.startswith("rust-"):
                name, version = self.fetch_name_version_from_nvr(nvr)
                name = self.fix_subpackage_name(name)
                if name[5:] != crate:
                    res.add((name[5:], version))
        return res

    def fix_subpackage_name(self, name):
        """
        Parse package name from a subpackage name
        Example: rust-cookie_store+default-devel to rust-cookie_store
                 rust-cookie_store0.12-devel
                 rust-cookie_store
        """
        if "+" in name:
            name = name.split("+")[0]
        else:
            if "-devel" in name:
                name = name.split("-devel")[0]
        return name

    def fix_compat_pkg_name(self, name, version):
        """
        Parse crate name from compat package name
        Example: crate_A0.22.0 to crate_A
                 crate_B0.2 to crate_B
        """
        while len(version) > 0:
            if version in name:
                return name[: -len(version)]
            version = version[:-1]
        return name

    def parse_require(self, require):
        """
        Parse a require format to normalized requires
        Example: (crate(syn/proc-macro) >= 1.0.48 with crate(syn/proc-macro) < 2.0.0~)
                  crate(enum-as-inner) = 0.3.4
        """
        pattern = r"crate\([/\w-]+\) [=><]+ \d+.\d+.\d+"
        if "with" in require:
            requires = [req for req in require.split("with")]
        else:
            requires = [require]
        normalized_reqs = []
        name = ""
        for req in requires:
            match = re.search(pattern, req)
            if not match:
                continue
            req = match.group(0)
            name, kind, version = req.split(" ")
            if "/" in name:
                name = name.split("/")[0][6:]
            else:
                name = name[6:-1]
            if "~" in version:
                version = version[:-1]
            normalized_reqs.append(
                (kind, CargoSemVer.coerce(CargoSemVer.parse(version).version))
            )
        return name, normalized_reqs

    def get_dependencies_from_repoquery(self, name):
        """Use dnf repoquery --requires to get all dependencies of a package"""
        cmd = f"sudo dnf repoquery --disablerepo=\* --enablerepo=rawhide,rawhide-source --requires rust-{name} --nvr"
        run = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        output, err = run.communicate()
        if run.returncode != 0:
            print(f"Fail to execute command {cmd}")
            print(err.decode())
            return set()
        res = {}
        # process all requires
        for require in output.decode().split("\n"):
            if "crate" in require:
                name, normalized_reqs = self.parse_require(require)
                res[name] = (require, normalized_reqs)
        return res

    def bodhi_check(self, bodhi_id: str):
        """
        Check compatibility for a Bodhi update
        """
        print(
            "=============================================================================================="
        )
        print(f"Update Compatibility Check for Bodhi Update {bodhi_id}")
        print(
            "============================================================================================="
        )
        res = self.bodhi_client.query(updateid=bodhi_id)
        if res["updates"] == []:
            print(f"Invalid Bodhi Update ID {bodhi_id}")
            return
        # initialize packages metadata for compatibility check
        # if a pkg is not in crate.io -> compat crate
        # if a package review bug is associated with a pkg -> new pkg
        builds = [build["nvr"] for build in res["updates"][0]["builds"]]
        bug_titles = [bug["title"] for bug in res["updates"][0]["bugs"]]
        pkg_reviews = set()
        for bug_title in bug_titles:
            pkg_reviews.add(self.fetch_new_pkg_from_bug(bug_title))
        updates = []
        new_pkgs = []
        pkgs = {}
        compat_pkgs = []
        for nvr in builds:
            if not nvr.startswith("rust-"):
                continue
            name, version = self.fetch_name_version_from_nvr(nvr)
            if self.is_compat_pkg(name[5:]):
                compat_pkgs.append((name, version, nvr))
            elif name in pkg_reviews:
                new_pkgs.append(nvr)
            else:
                updates.append(nvr)
                pkgs[name] = {"type": "existing", "version": version}

        # print builds information of this se
        print("The update set contains following builds:")
        for nvr in new_pkgs:
            print(f"      [New]    {nvr}")
        for nvr in updates:
            print(f"      [Update] {nvr}")
        for _, _, nvr in compat_pkgs:
            print(f"      [Compat] {nvr}")

        compat_found, _ = self.check_compatibility(pkgs)
        if len(compat_pkgs) + len(compat_found) == 0:
            return
        print(
            "=============================================================================================="
        )
        print(f"Compat Crates Found")
        print(
            "============================================================================================="
        )

        if len(compat_pkgs) > 0:
            print("Compat packages found in the same update set:")
            for name, version, _ in compat_pkgs:
                print(f"      {name} {version}")

        if len(compat_found) > 0:
            print("Compat packages found when doing repo query:")
            for name, version in sorted(list(compat_found)):
                print(f"      {name} {version}")
