# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import subprocess
import json

from .rust_update import CHECKOUT_METADATA_FILE

DIVIDING_LINE = "="*150

def print_cmd_output(output, prefix=""):
        """ Print cmd output with warning color"""
        out = output.stdout.decode().split("\n")[:-1]
        err = output.stderr.decode().split("\n")[:-1]
        for line in out:
            print('\033[93m'+prefix+line+'\033[0m')
        for line in err:
            print('\033[93m'+prefix+line+'\033[0m')
        return len(out)+len(err)

def dump_metadata(checkout_metadata):
    """ Dump all checkout related metadata into json file"""
    with open(CHECKOUT_METADATA_FILE, "w") as f:
        json.dump(checkout_metadata, f, indent=4)

class Push:

    @staticmethod
    def push_repo(pkg, version, checkout_metadata):
        """ Fork repository and push changes"""
        os.chdir(f"{os.getcwd()}/{pkg}")
        print(f"Push to {pkg} repository")
        # git add .
        output = subprocess.run(["git", "add", "."], capture_output=True)
        if output.returncode != 0:
            print('\033[93m'+"     Fail to execute commad 'git add .'"+'\033[0m')
            print_cmd_output(output, prefix=" "*6)
            os.chdir(f"{os.getcwd()}/..")
            return False
        # git commit -m "Update to {version}"
        output = subprocess.run(["git", "commit", "-m", f"Update to {version}"], capture_output=True)
        if output.returncode != 0:
            print('\033[93m'+f"      Fail to execute commad \"git commit -m 'Update to {version}'\" .'"+'\033[0m')
            print_cmd_output(output, prefix=" "*6)
            os.chdir(f"{os.getcwd()}/..")
            return False
        print(f"      Create Commit 'Update to {version}'")
        # try push to actual release repository first, if fails, do fork
        # git push
        output = subprocess.run(["git", "push"], capture_output=True)
        if output.returncode != 0:
            print('\033[93m'+f"      Fail to push to upstream repository"+'\033[0m')
            print_cmd_output(output, prefix=" "*6)
        else:
            print(f"      Push update to https://src.fedoraproject.org/rpms/{pkg}")
            checkout_metadata["checkout"][pkg]["push"] = "origin"
            os.chdir(f"{os.getcwd()}/..")
            dump_metadata(checkout_metadata)
            return True
        # fedpkg fork
        output = subprocess.run(["fedpkg", "fork"], capture_output=True)
        if output.returncode == 0:
            out = output.stdout.decode().split("\n")
            name = ""
            for line in out:
                if line.startswith("Adding as remote ") or line.startswith("Remote with name "):
                    name = line.split("'")[1].split("'")[0]
            print(f"      Fork {pkg}: https://src.fedoraproject.org/fork/{name}/rpms/{pkg}")
        else:
            print_cmd_output(output, prefix=" "*6)
            os.chdir(f"{os.getcwd()}/..")
            return False
        # git push username
        output = subprocess.run(["git", "push", name], capture_output=True)
        if output.returncode != 0:
            print('\033[93m'+f"      Fail to execute commad 'git push {name}'"+'\033[0m')
            print_cmd_output(output, prefix=" "*6)
            os.chdir(f"{os.getcwd()}/..")
            return False
        checkout_metadata["checkout"][pkg]["push"] = name
        print(f"      Push update to https://src.fedoraproject.org/fork/{name}/rpms/{pkg}")
        print(f"      To create a pull-request for upstream repo: https://src.fedoraproject.org/fork/{name}/rpms/{pkg}/diff/rawhide..rawhide")
        os.chdir(f"{os.getcwd()}/..")
        dump_metadata(checkout_metadata)
        return True

    @staticmethod
    def push(checkout_metadata):
        """ Push changes for exisint packages """
        pkgs = checkout_metadata["checkout"]
        print(DIVIDING_LINE)
        print("Push")
        print(DIVIDING_LINE)
        push_fails = []
        for pkg in pkgs:
            if pkgs[pkg]["type"] == "existing":
                if not Push.push_repo(pkg, pkgs[pkg]["version"], checkout_metadata):
                    push_fails.append(pkg)
        print("-"*150)
        if len(push_fails) != 0:
            print("Fail to push following repos, please check:")
            for pkg in push_fails:
                print(f"      {pkg}")
        else:
            print("All changes are pushed to forked repos.")

    @staticmethod
    def push_branch(pkgs, branch):
        """ Push release branch for all packages """
        print(DIVIDING_LINE)
        print(f"Push Release Branch {branch}")
        print(DIVIDING_LINE)
        for pkg in pkgs:
            pkg_info = pkgs[pkg]
            if pkg_info["type"] == "existing":
                # git switch branch
                print(f"Push branch {branch} for {pkg}")
                os.chdir(f"{os.getcwd()}/{pkg}")
                output = subprocess.run(["git", "switch", branch], capture_output=True)
                if output.returncode == 0:
                    print(f"      Successfully switch branch to {branch}")
                else:
                    print(f"      Failed to switch branch to {branch}")
                    print_cmd_output(output, prefix=6*" ")
                    os.chdir(f"{os.getcwd()}/..")
                    continue
                # git push
                output = subprocess.run(["git", "push"], capture_output=True)
                if output.returncode == 0:
                    print(f"      Successfully push branch {branch}")
                    subprocess.run(["git", "switch", "rawhide"], capture_output=True)
                    os.chdir(f"{os.getcwd()}/..")
                    continue
                else:
                    # if git push fails, do fork and prompt to create pull-request
                    # Fork should be done when working with rawhide, just do it again here to get the correct remote name
                    print(f"      Failed to push branch {branch} to upstream repository")
                    print_cmd_output(output, prefix=6*" ")
                    # fork and push repo, prompt to create pull-request
                    output = subprocess.run(["fedpkg", "fork"], capture_output=True)
                    if output.returncode == 0:
                        out = output.stdout.decode().split("\n")
                        name = ""
                        for line in out:
                            if line.startswith("Adding as remote ") or line.startswith("Remote with name "):
                                name = line.split("'")[1].split("'")[0]
                        print(f"Fork {pkg}: https://src.fedoraproject.org/fork/{name}/rpms/{pkg}")
                    else:
                        print('\033[93m'+f"Fail to fetch forked repo remote name"+'\033[0m')
                        print_cmd_output(output)
                        os.chdir(f"{os.getcwd()}/..")
                        continue
                    # git push username
                    output = subprocess.run(["git", "push", name], capture_output=True)
                    if output.returncode != 0:
                        print('\033[93m'+f"Fail to execute commad 'git push {name}'"+'\033[0m')
                        print_cmd_output(output)
                        os.chdir(f"{os.getcwd()}/..")
                        continue
                    print(f"Push update to https://src.fedoraproject.org/fork/{name}/rpms/{pkg}")
                    print(f"To create a pull-request for upstream repo: https://src.fedoraproject.org/fork/{name}/rpms/{pkg}/diff/{branch}..{branch}")
                os.chdir(f"{os.getcwd()}/..")
