# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import subprocess

def generate_spec(crate, version, patch=False):
    """ Generate a new spec file using rust2rpm, used for new packages"""
    if patch:
        output = subprocess.run(["rust2rpm", "-s", "-p", crate, version])
    else:
        output = subprocess.run(["rust2rpm", "-s", crate, version], capture_output=True)
    if output.returncode != 0:
        print(f"Fail to generate new spec for {crate} with version {version}")
        print_cmd_output(output)
        return
    else:
        print(f"Create new spec for {crate} with version {version}")
    output = subprocess.run(["cp", f"rust-{crate}.spec", f"rust-{crate}.unmodified.spec"], capture_output=True)

def update_spec(crate, new_version, patch=False):
    """
    Update spec file using rust2rpm, preserve changes from old spec file iff .unmodified.spec is present
    1. Move .spec and .unmodified.spec to .old.modified.spec and .old.unmodified.spec 
    2. Create new .spec and a copy .unmodified.spec using rust2rpm
    3. Create patch between .old.unmodified.spec and .old.modified.spec
    4. Apply patch to new .spec
    """
    # rename current spec to .modified.spec and .unmodified.spec to .old.unmodified.spec
    patch_from_old_spec = False
    if os.path.exists(f"rust-{crate}.spec") and os.path.exists(f"rust-{crate}.unmodified.spec"):
        patch_from_old_spec = True
        output = subprocess.run(["cp", f"rust-{crate}.spec", f"rust-{crate}.old.modified.spec"], capture_output=True)
        output = subprocess.run(["mv", f"rust-{crate}.unmodified.spec", f"rust-{crate}.old.unmodified.spec"], capture_output=True)

    # generate new spec file with patch option, if patch==False, add the existing patch to new spec
    generate_spec(crate=crate, version=new_version, patch=patch)

    # remove {crate}-fix-data.diff file if it is an old one
    if os.path.exists(f"{crate}-fix-metadata.diff"):
        is_old = False
        with open(f"{crate}-fix-metadata.diff") as f:
            line = f.read()
            if f"{crate}-{new_version}/Cargo.toml" not in line:
                is_old = True
        if is_old:
            output = subprocess.run(["rm", f"{crate}-fix-metadata.diff"], capture_output=True)
    
    # remove {crate}-fix-data-auto.diff file if it is an old one
    if os.path.exists(f"{crate}-fix-metadata-auto.diff"):
        is_old = False
        with open(f"{crate}-fix-metadata-auto.diff") as f:
            line = f.read()
            if f"{crate}-{new_version}/Cargo.toml" not in line:
                is_old = True
        if is_old:
            output = subprocess.run(["rm", f"{crate}-fix-metadata-auto.diff"], capture_output=True)

    if not patch_from_old_spec:
        return False
    # sync some info in old.modified and old.unmodified spec to avoid patch conflicts
    # 1. rust2rpm version line
    # 2. crate version should be the same as new spec
    # 3. crate summary should be the same as new spec
    head = summary = ""
    version_line = f"Version:        {new_version}\n"
    with open(f"rust-{crate}.spec") as f:
        cont = f.readlines()
        head = cont[0]
        for line in cont:
            if line.startswith("Summary:"):
                summary = line
                break
    with open(f"rust-{crate}.old.modified.spec", "r+") as fmod, open(f"rust-{crate}.old.unmodified.spec", "r+") as fumod:
        modified_cont = fmod.readlines()
        unmodified_cont = fumod.readlines()
        # avoid conflict caused by rust2rpm version line
        modified_cont[0] = unmodified_cont[0] = head 
        # avoid conflict caused by version and summary
        updated = False
        for i, line in enumerate(modified_cont):
            if line.startswith("Version:"):
                modified_cont[i] = version_line
            if line.startswith("Summary:") and not updated:
                modified_cont[i] = summary
                updated = True
                break
        updated = False
        for i, line in enumerate(unmodified_cont):
            if line.startswith("Version:"):
                unmodified_cont[i] = version_line
            if line.startswith("Summary:") and not updated:
                modified_cont[i] = summary
                updated = True
                break
        fmod.seek(0)
        fumod.seek(0)
        fmod.writelines(modified_cont)
        fumod.writelines(unmodified_cont)

    # Create spec.patch, patch it to new spec file and remove spec.patch file
    patch_warning = False
    if not os.path.exists("spec.patch"): output = subprocess.check_output(["touch", "spec.patch"])
    with open("spec.patch", "r+") as patch:
        output = subprocess.run(["diff", "-u", f"rust-{crate}.old.unmodified.spec", f"rust-{crate}.old.modified.spec"], stdout=patch)
        output = subprocess.run(["rm", f"rust-{crate}.old.unmodified.spec", f"rust-{crate}.old.modified.spec"], capture_output=True)
        patch.seek(0)
        output = subprocess.run(["patch", f"rust-{crate}.spec"], stdin=patch, capture_output=True)
        if output.returncode != 0:
            print_cmd_output(output)
            patch_warning = True
        else:
            n_lines = print_cmd_output(output)
            if n_lines > 1:
                patch_warning = True
    if os.path.exists(".gitignore"):
        with open(".gitignore", "r+") as ignore:
            for line in ignore.readlines():
                if "spec.patch" in line:
                    return
            ignore.write("spec.patch\n")
    output = subprocess.run(["rm", f"rust-{crate}.spec"], capture_output=True)
    return patch_warning

def print_cmd_output(output):
        """ Print cmd output with warning color"""
        out = output.stdout.decode().split("\n")[:-1]
        err = output.stderr.decode().split("\n")[:-1]
        for line in out:
            print('\033[93m'+line+'\033[0m')
        for line in err:
            print('\033[93m'+line+'\033[0m')
        return len(out)+len(err)