# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import sys
from typing import List, Tuple
sys.path.append("..")
from rust_update_set.dependency_check import DChecker
from dataclasses import dataclass
from unittest import mock
from rust2rpm.core.metadata import Version

class DependencyCheckTest(unittest.TestCase):

    def test_parse_require(self):

        @dataclass
        class Test_case:
            require: str
            name: str
            normalized_reqs: List[Tuple[str, Version]]

        test_cases = [
            Test_case(
                require="(crate(syn/proc-macro) >= 1.0.48 with crate(syn/proc-macro) < 2.0.0~)",
                name = "syn",
                normalized_reqs=[(">=", Version(1, 0, 48, None, None)), ("<", Version(2, 0, 0, None, None))]
            ),
            Test_case(require="crate(enum-as-inner) = 0.3.4", 
                name="enum-as-inner", 
                normalized_reqs=[("=", Version(0, 3, 4, None, None))]
            )
        ]

        dc = DChecker(bodhi_client=mock.Mock())
        for test_case in test_cases:
            name, normalized_reqs = dc.parse_require(test_case.require)
            self.assertEqual((name, normalized_reqs), (test_case.name, test_case.normalized_reqs))
        
    def test_fix_compat_pkg_name(self):

        @dataclass
        class Test_case:
            name: str
            version: str
            fixed_name: str

        test_cases = [
            Test_case(
                name="crateA0.22.22",
                version="0.22.22",
                fixed_name="crateA"
            ),
            Test_case(
                name="crateB1.2",
                version="1.2.3",
                fixed_name="crateB"
            ),
            Test_case(
                name="crateC2",
                version="2.0.1",
                fixed_name="crateC"
            ),
        ]

        dc = DChecker(bodhi_client=mock.Mock())
        for test_case in test_cases:
            self.assertEqual(dc.fix_compat_pkg_name(test_case.name, test_case.version), test_case.fixed_name)

    def test_fix_subpackage_name(self):

        @dataclass
        class Test_case:
            subpkg_name: str
            pkg_name: str

        test_cases = [
            Test_case(
                subpkg_name="rust-cookie+default-devel",
                pkg_name="rust-cookie"
            ),
            Test_case(
                subpkg_name="rust-fruit+devel-devel",
                pkg_name="rust-fruit"
            ),
            Test_case(
                subpkg_name="rust-ice_cream0.12-devel",
                pkg_name="rust-ice_cream0.12"
            ),
            Test_case(
                subpkg_name="rust-burger",
                pkg_name="rust-burger"
            ),
        ]

        dc = DChecker(bodhi_client=mock.Mock())
        for test_case in test_cases:
            self.assertEqual(dc.fix_subpackage_name(test_case.subpkg_name), test_case.pkg_name)

    def test_fetch_name_version_from_nvr(self):

        @dataclass
        class Test_case:
            nvr: str
            name: str
            version: str

        test_cases = [
            Test_case(
                nvr="rust-sysinfo+debug-devel-0.19.2-2.fc36",
                name="rust-sysinfo+debug-devel",
                version="0.19.2"
            ),
            Test_case(
                nvr="rust-ring-0.16.20-4.fc37",
                name="rust-ring",
                version="0.16.20"
            ),
            Test_case(
                nvr="rust-gmp-mpfr-sys-devel-1.4.9-1.fc37",
                name="rust-gmp-mpfr-sys-devel",
                version="1.4.9"
            ),
        ]

        dc = DChecker(bodhi_client=mock.Mock())
        for test_case in test_cases:
            self.assertEqual((dc.fetch_name_version_from_nvr(test_case.nvr)), (test_case.name, test_case.version))

    def test_fetch_new_pkg_from_bug(self):

        @dataclass
        class Test_case:
            title: str
            pkg: str

        test_cases = [
            Test_case(
                title="Review Request: rust-pakcageA",
                pkg="pakcageA",
            ),
            Test_case(
                title="Review Request: rust-pakcageB-lib",
                pkg="pakcageB-lib",
            ),
            Test_case(
                title="Review Request: rust-pakcageC_lib",
                pkg="pakcageC_lib",
            ),
        ]

        dc = DChecker(bodhi_client=mock.Mock())
        for test_case in test_cases:
            self.assertEqual(dc.fetch_new_pkg_from_bug(test_case.title), test_case.pkg)

if __name__ == "__main__":
    unittest.main()
